import inspect
from typing import Any, ForwardRef, Optional, Tuple, Type, Union

from pydantic import BaseConfig
from pydantic.fields import (
    SHAPE_FROZENSET,
    SHAPE_LIST,
    SHAPE_SEQUENCE,
    SHAPE_SET,
    SHAPE_TUPLE,
    SHAPE_TUPLE_ELLIPSIS,
    FieldInfo,
    ModelField,
    Required,
    Undefined,
    UndefinedType,
)
from pydantic.schema import get_annotation_from_field_info
from pydantic.typing import evaluate_forwardref, get_args, get_origin
from typing_extensions import Annotated

from fast_depends import model
from fast_depends.library import CustomField
from fast_depends.types import AnyCallable, AnyDict
from fast_depends.utils import is_coroutine_callable

sequence_shapes = {
    SHAPE_LIST,
    SHAPE_SET,
    SHAPE_FROZENSET,
    SHAPE_TUPLE,
    SHAPE_SEQUENCE,
    SHAPE_TUPLE_ELLIPSIS,
}
sequence_types = (list, set, tuple)


def get_dependant(
    *,
    path: str,
    call: AnyCallable,
    name: Optional[str] = None,
    use_cache: bool = True,
) -> model.Dependant:
    dependant = model.Dependant(
        call=call,
        path=path,
        name=name,
        use_cache=use_cache,
        return_field=None,
    )

    is_async = is_coroutine_callable(call)

    endpoint_signature = get_typed_signature(call)
    signature_params = endpoint_signature.parameters

    for param in signature_params.values():
        custom, depends, param_field = analyze_param(
            param_name=param.name,
            annotation=param.annotation,
            default=param.default,
        )

        if param.name == model.RETURN_FIELD:
            dependant.return_field = param_field
            continue

        elif custom is not None:
            dependant.custom.append(custom)

        elif depends is not None:
            assert is_async or not is_coroutine_callable(
                depends.dependency
            ), f"You cannot use async dependency `{depends}` with sync `{dependant}`"

            sub_dependant = get_param_sub_dependant(
                param_name=param.name,
                depends=depends,
                path=path,
            )
            dependant.dependencies.append(sub_dependant)

        dependant.params.append(param_field)

    return dependant


def analyze_param(
    *,
    param_name: str,
    annotation: Any,
    default: Any,
) -> Tuple[Optional[CustomField], Optional[model.Depends], ModelField]:
    depends = None
    custom = None
    field_info = None

    if (
        annotation is not inspect.Signature.empty
        and get_origin(annotation) is Annotated  # type: ignore[comparison-overlap]
    ):
        annotated_args = get_args(annotation)
        custom_annotations = [
            arg
            for arg in annotated_args[1:]
            if isinstance(arg, (FieldInfo, model.Depends, CustomField))
        ]

        custom_annotations = next(iter(custom_annotations), None)
        if isinstance(custom_annotations, FieldInfo):
            field_info = custom_annotations
            assert field_info.default is Undefined or field_info.default is Required, (
                f"`{field_info.__class__.__name__}` default value cannot be set in"
                f" `Annotated` for {param_name!r}. Set the default value with `=` instead."
            )
            field_info.default = Required

        elif isinstance(custom_annotations, model.Depends):
            depends = custom_annotations

        elif isinstance(custom_annotations, CustomField):  # pragma: no branch
            custom_annotations.set_param_name(param_name)
            custom = custom_annotations
            if custom.cast is False:
                annotation = Any

    if isinstance(default, model.Depends):
        assert depends is None, (
            "Cannot specify `Depends` in `Annotated` and default value"
            f" together for {param_name!r}"
        )
        assert field_info is None, (
            "Cannot specify a annotation in `Annotated` and `Depends` as a"
            f" default value together for {param_name!r}"
        )
        depends = default

    elif isinstance(default, CustomField):
        default.set_param_name(param_name)
        custom = default
        if custom.cast is False:
            annotation = Any

    elif isinstance(default, FieldInfo):
        assert field_info is None, (
            "Cannot specify annotations in `Annotated` and default value"
            f" together for {param_name!r}"
        )
        field_info = default

    if (depends or custom) is not None:
        field = None

    if field_info is not None:
        annotation = get_annotation_from_field_info(
            annotation if annotation is not inspect.Signature.empty else Any,
            field_info,
            param_name,
        )
    else:
        field_info = FieldInfo(default=default)

    alias = field_info.alias or param_name

    if custom and custom.required is True:
        required = True
    else:
        required = field_info.default in (Required, Undefined, inspect._empty)

    field = create_response_field(
        name=param_name,
        type_=Any if depends and depends.cast is False else annotation,
        default=None if any((depends, custom)) else field_info.default,
        alias=alias,
        required=required,
        field_info=field_info,
    )

    return custom, depends, field


def get_typed_signature(call: AnyCallable) -> inspect.Signature:
    signature = inspect.signature(call)
    globalns = getattr(call, "__globals__", {})
    typed_params = [
        inspect.Parameter(
            name=param.name,
            kind=param.kind,
            default=param.default,
            annotation=get_typed_annotation(param.annotation, globalns),
        )
        for param in signature.parameters.values()
    ]

    if signature.return_annotation is not signature.empty:
        typed_params.append(
            inspect.Parameter(
                name=model.RETURN_FIELD,
                kind=inspect._KEYWORD_ONLY,
                annotation=get_typed_annotation(signature.return_annotation, globalns),
            )
        )
    typed_signature = inspect.Signature(typed_params)
    return typed_signature


def get_typed_annotation(annotation: Any, globalns: AnyDict) -> Any:
    if isinstance(annotation, str):
        try:
            annotation = ForwardRef(annotation)
            annotation = evaluate_forwardref(annotation, globalns, globalns)
        except Exception:
            raise ValueError(  # noqa: B904
                f"Invalid filed annotation! Hint: check that {annotation} is a valid pydantic field type"
            )
    return annotation


def get_param_sub_dependant(
    *,
    param_name: str,
    depends: model.Depends,
    path: str,
) -> model.Dependant:
    assert depends.dependency
    return get_sub_dependant(
        depends=depends,
        dependency=depends.dependency,
        path=path,
        name=param_name,
    )


def get_sub_dependant(
    *,
    depends: model.Depends,
    dependency: AnyCallable,
    path: str,
    name: Optional[str] = None,
) -> model.Dependant:
    sub_dependant = get_dependant(
        path=path,
        call=dependency,
        name=name,
        use_cache=depends.use_cache,
    )
    return sub_dependant


def create_response_field(
    name: str,
    type_: Type[Any],
    default: Optional[Any] = None,
    required: Union[bool, UndefinedType] = True,
    field_info: Optional[FieldInfo] = None,
    alias: Optional[str] = None,
) -> ModelField:
    """
    Create a new response field. Raises if type_ is invalid.
    """
    try:
        return ModelField(
            name=name,
            type_=type_ if type_ is not inspect._empty else Any,
            default=default,
            required=required,
            class_validators={},
            model_config=BaseConfig,
            alias=alias,
            field_info=field_info or FieldInfo(),
        )
    except RuntimeError:  # pragma: no cover
        raise ValueError(  # noqa: B904
            f"Invalid args for response field! Hint: check that {type_} is a valid pydantic field type"
        )



from contextlib import AsyncExitStack, ExitStack
from copy import deepcopy
from typing import Any, Dict, List, Mapping, Optional, Sequence, Tuple, TypeVar, cast

from pydantic.error_wrappers import ErrorList, ErrorWrapper
from pydantic.errors import MissingError
from pydantic.fields import ModelField

from fast_depends.construct import get_dependant
from fast_depends.model import Dependant
from fast_depends.types import AnyCallable, AnyDict
from fast_depends.utils import (
    is_async_gen_callable,
    is_coroutine_callable,
    is_gen_callable,
    run_async,
    solve_generator_async,
    solve_generator_sync,
)

T = TypeVar("T")


async def solve_dependencies_async(
    *,
    dependant: Dependant,
    stack: AsyncExitStack,
    body: Optional[AnyDict] = None,
    dependency_overrides_provider: Optional[Any] = None,
    dependency_cache: Optional[Dict[Tuple[AnyCallable, Tuple[str]], Any]] = None,
) -> Tuple[Dict[str, Any], List[ErrorList], Dict[Tuple[AnyCallable, Tuple[str]], Any],]:
    errors: List[ErrorList] = []

    dependency_cache = dependency_cache or {}

    sub_dependant: Dependant
    for sub_dependant in dependant.dependencies:
        sub_dependant.call = cast(AnyCallable, sub_dependant.call)
        sub_dependant.cache_key = cast(
            Tuple[AnyCallable, Tuple[str]], sub_dependant.cache_key
        )
        call = sub_dependant.call
        use_sub_dependant = sub_dependant
        if (
            dependency_overrides_provider
            and dependency_overrides_provider.dependency_overrides
        ):
            call = getattr(
                dependency_overrides_provider, "dependency_overrides", {}
            ).get(sub_dependant.call)
            if call is not None:  # pragma: no branch
                use_sub_dependant = get_dependant(
                    path=sub_dependant.path,
                    call=call,
                    name=sub_dependant.name,
                )
        assert call
        solved_result = await solve_dependencies_async(
            dependant=use_sub_dependant,
            body=body,
            dependency_overrides_provider=dependency_overrides_provider,
            dependency_cache=dependency_cache,
            stack=stack,
        )
        (
            sub_values,
            sub_errors,
            sub_dependency_cache,
        ) = solved_result

        dependency_cache.update(sub_dependency_cache)

        if sub_errors:
            errors.extend(sub_errors)
            continue

        if (
            use_sub_dependant.use_cache
            and use_sub_dependant.cache_key in dependency_cache
        ):
            solved = dependency_cache[use_sub_dependant.cache_key]
        elif is_gen_callable(call) or is_async_gen_callable(call):
            solved, sub_errors = use_sub_dependant.cast_response(
                await solve_generator_async(
                    call=call, stack=stack, sub_values=sub_values
                )
            )
        else:
            solved, sub_errors = use_sub_dependant.cast_response(
                await run_async(use_sub_dependant.call, **sub_values)
            )

        if sub_errors:
            errors.append(sub_errors)
            continue

        if use_sub_dependant.name is not None:  # pragma: no branch
            body[use_sub_dependant.name] = solved

        if use_sub_dependant.cache_key not in dependency_cache:
            dependency_cache[use_sub_dependant.cache_key] = solved

    for custom in dependant.custom:
        body = await run_async(custom.use, **(body or {}))

    params, main_errors = params_to_args(dependant.params, body or {})
    errors.extend(main_errors)
    return params, errors, dependency_cache


def solve_dependencies_sync(
    *,
    dependant: Dependant,
    stack: ExitStack,
    body: Optional[AnyDict] = None,
    dependency_overrides_provider: Optional[Any] = None,
    dependency_cache: Optional[Dict[Tuple[AnyCallable, Tuple[str]], Any]] = None,
) -> Tuple[Dict[str, Any], List[ErrorList], Dict[Tuple[AnyCallable, Tuple[str]], Any],]:
    assert not is_coroutine_callable(dependant.call) and not is_async_gen_callable(
        dependant.call
    ), f"You can't call async `{dependant.call.__name__}` at sync code"

    errors: List[ErrorList] = []

    dependency_cache = dependency_cache or {}

    sub_dependant: Dependant
    for sub_dependant in dependant.dependencies:
        sub_dependant.call = cast(AnyCallable, sub_dependant.call)
        sub_dependant.cache_key = cast(
            Tuple[AnyCallable, Tuple[str]], sub_dependant.cache_key
        )
        call = sub_dependant.call
        use_sub_dependant = sub_dependant
        if (
            dependency_overrides_provider
            and dependency_overrides_provider.dependency_overrides
        ):
            call = getattr(
                dependency_overrides_provider, "dependency_overrides", {}
            ).get(sub_dependant.call)
            if call is not None:  # pragma: no branch
                use_sub_dependant = get_dependant(
                    path=sub_dependant.path,
                    call=call,
                    name=sub_dependant.name,
                )
        assert call
        solved_result = solve_dependencies_sync(
            dependant=use_sub_dependant,
            body=body,
            stack=stack,
            dependency_overrides_provider=dependency_overrides_provider,
            dependency_cache=dependency_cache,
        )
        (
            sub_values,
            sub_errors,
            sub_dependency_cache,
        ) = solved_result

        dependency_cache.update(sub_dependency_cache)

        if sub_errors:
            errors.extend(sub_errors)
            continue

        if (
            use_sub_dependant.use_cache
            and use_sub_dependant.cache_key in dependency_cache
        ):
            solved = dependency_cache[sub_dependant.cache_key]
        elif is_gen_callable(call):
            solved, sub_errors = use_sub_dependant.cast_response(
                solve_generator_sync(call=call, stack=stack, sub_values=sub_values)
            )
        else:
            solved, sub_errors = use_sub_dependant.cast_response(call(**sub_values))

        if sub_errors:
            errors.append(sub_errors)
            continue

        if use_sub_dependant.name is not None:  # pragma: no branch
            body[sub_dependant.name] = solved

        if use_sub_dependant.cache_key not in dependency_cache:
            dependency_cache[use_sub_dependant.cache_key] = solved

    for custom in dependant.custom:
        assert not is_coroutine_callable(custom.use) and not is_async_gen_callable(
            custom.use
        ), f"You can't use async `{type(custom).__name__}` at sync code"
        body = custom.use(**(body or {}))

    params, main_errors = params_to_args(dependant.params, body or {})
    errors.extend(main_errors)
    return params, errors, dependency_cache


def params_to_args(
    required_params: Sequence[ModelField],
    received_params: Mapping[str, Any],
) -> Tuple[AnyDict, List[ErrorList]]:
    values: AnyDict = {}
    errors: List[ErrorList] = []
    for field in required_params:
        value = received_params.get(field.alias)
        if value is None:
            if field.required:
                errors.append(ErrorWrapper(MissingError(), loc=(field.alias,)))
            else:
                values[field.name] = deepcopy(field.default)
            continue

        v_, errors_ = field.validate(value, values, loc=(field.alias,))
        if isinstance(errors_, ErrorWrapper):
            errors.append(errors_)
        elif isinstance(errors_, list):  # pragma: no cover
            errors.extend(errors_)
        else:
            values[field.name] = v_
    return values, errors



from contextlib import AsyncExitStack, ExitStack
from functools import partial, wraps
from typing import Any, Callable, Optional, TypeVar

from pydantic import ValidationError

from fast_depends import model
from fast_depends.construct import get_dependant
from fast_depends.injector import solve_dependencies_async, solve_dependencies_sync
from fast_depends.provider import dependency_provider
from fast_depends.types import AnyCallable, P
from fast_depends.utils import args_to_kwargs, is_coroutine_callable, run_async

T = TypeVar("T")


def Depends(
    dependency: AnyCallable,
    *,
    use_cache: bool = True,
    cast: bool = True,
) -> Any:  # noqa: N802
    return model.Depends(dependency=dependency, use_cache=use_cache, cast=cast)


def wrap_dependant(dependant: model.Dependant) -> model.Dependant:
    return dependant


def inject(
    func: Callable[P, T],
    *,
    dependency_overrides_provider: Optional[Any] = dependency_provider,
    wrap_dependant: Callable[[model.Dependant], model.Dependant] = wrap_dependant,
) -> Callable[P, T]:
    dependant = get_dependant(call=func, path=func.__name__)

    dependant = wrap_dependant(dependant)

    if is_coroutine_callable(func) is True:
        f = async_typed_wrapper
    else:
        f = sync_typed_wrapper

    return wraps(func)(
        partial(
            f,
            dependant=dependant,
            dependency_overrides_provider=dependency_overrides_provider,
        )
    )


async def async_typed_wrapper(
    *args: P.args,
    dependant: model.Dependant,
    dependency_overrides_provider: Optional[Any],
    **kwargs: P.kwargs,
) -> Any:
    kwargs = args_to_kwargs((x.name for x in dependant.params), *args, **kwargs)

    async with AsyncExitStack() as stack:
        solved_result, errors, _ = await solve_dependencies_async(
            body=kwargs,
            dependant=dependant,
            stack=stack,
            dependency_overrides_provider=dependency_overrides_provider,
        )

        if errors:
            raise ValidationError(errors, dependant.error_model)

        v, casted_errors = dependant.cast_response(
            await run_async(dependant.call, **solved_result)
        )

        if casted_errors:
            raise ValidationError(errors, dependant.error_model)

        return v


def sync_typed_wrapper(
    *args: P.args,
    dependant: model.Dependant,
    dependency_overrides_provider: Optional[Any],
    **kwargs: P.kwargs,
) -> Any:
    kwargs = args_to_kwargs((x.name for x in dependant.params), *args, **kwargs)

    with ExitStack() as stack:
        solved_result, errors, _ = solve_dependencies_sync(
            body=kwargs,
            dependant=dependant,
            stack=stack,
            dependency_overrides_provider=dependency_overrides_provider,
        )

        if errors:
            raise ValidationError(errors, dependant.error_model)

        v, casted_errors = dependant.cast_response(dependant.call(**solved_result))

        if casted_errors:
            raise ValidationError(errors, dependant.error_model)

        return v



from typing import Dict

from fast_depends.types import AnyCallable


class Provider:
    def __init__(self) -> None:
        self.dependency_overrides: Dict[AnyCallable, AnyCallable] = {}

    def override(self, original: AnyCallable, override: AnyCallable) -> None:
        self.dependency_overrides[original] = override

    def clear(self) -> None:
        self.dependency_overrides = {}


dependency_provider = Provider()



from typing import Any, List, Optional, Tuple, List

from pydantic import create_model
from pydantic.error_wrappers import ErrorList
from pydantic.fields import ModelField

from fast_depends.library import CustomField
from fast_depends.types import AnyCallable

RETURN_FIELD = "custom_return"


class Dependant:
    def __init__(
        self,
        *,
        call: AnyCallable,
        params: Optional[List[ModelField]] = None,
        return_field: Optional[ModelField] = None,
        dependencies: Optional[List["Dependant"]] = None,
        custom: Optional[List[CustomField]] = None,
        use_cache: bool = True,
        path: Optional[str] = None,
        name: Optional[str] = None,
    ) -> None:
        self.params = params or []
        self.return_field = return_field
        self.dependencies = dependencies or []
        self.custom = custom or []
        self.call = call
        self.use_cache = use_cache
        # Parent argument name at subdependency
        self.name = name
        # Store the path to be able to re-generate a dependable from it in overrides
        self.path = path
        # Save the cache key at creation to optimize performance
        self.cache_key = (self.call,)
        self.error_model = create_model(getattr(call, "__name__", str(call)))

    @property
    def real_params(self) -> List[ModelField]:
        custom = tuple(c.param_name for c in self.custom)
        return list(filter(lambda x: x.name not in custom, self.params))

    @property
    def flat_params(self) -> List[ModelField]:
        params = self.real_params
        for d in self.dependencies:
            params.extend(d.flat_params)

        params_unique = []
        for p in params:
            if p not in params_unique:
                params_unique.append(p)

        return params_unique

    def cast_response(self, response: Any) -> Tuple[Optional[Any], Optional[ErrorList]]:
        if self.return_field is None:
            return response, []
        return self.return_field.validate(response, {}, loc=RETURN_FIELD)


class Depends:
    def __init__(
        self,
        dependency: AnyCallable,
        *,
        use_cache: bool = True,
        cast: bool = True,
    ) -> None:
        self.dependency = dependency
        self.use_cache = use_cache
        self.cast = cast

    def __repr__(self) -> str:
        attr = getattr(self.dependency, "__name__", type(self.dependency).__name__)
        cache = "" if self.use_cache else ", use_cache=False"
        return f"{self.__class__.__name__}({attr}{cache})"



import asyncio
import functools
import inspect
from contextlib import AsyncExitStack, ExitStack, asynccontextmanager, contextmanager
from typing import (
    Any,
    AsyncGenerator,
    Callable,
    ContextManager,
    Dict,
    Iterable,
    TypeVar,
    cast,
)

import anyio

from fast_depends.types import AnyCallable, AnyDict, P

T = TypeVar("T")


def args_to_kwargs(
    arguments: Iterable[str], *args: P.args, **kwargs: P.kwargs
) -> AnyDict:
    if not args:
        return kwargs

    unused = filter(lambda x: x not in kwargs, arguments)

    return dict((*zip(unused, args), *kwargs.items()))


async def run_async(func: Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
    if asyncio.iscoroutinefunction(func):
        r = await func(*args, **kwargs)
        return cast(T, r)
    else:
        return await run_in_threadpool(func, *args, **kwargs)


async def run_in_threadpool(
    func: Callable[P, T], *args: P.args, **kwargs: P.kwargs
) -> T:
    if kwargs:  # pragma: no cover
        func = functools.partial(func, **kwargs)
    return await anyio.to_thread.run_sync(func, *args)


def is_async_gen_callable(call: AnyCallable) -> bool:
    if inspect.isasyncgenfunction(call):
        return True
    dunder_call = getattr(call, "__call__", None)  # noqa: B004
    return inspect.isasyncgenfunction(dunder_call)


def is_gen_callable(call: AnyCallable) -> bool:
    if inspect.isgeneratorfunction(call):
        return True
    dunder_call = getattr(call, "__call__", None)  # noqa: B004
    return inspect.isgeneratorfunction(dunder_call)


def is_coroutine_callable(call: AnyCallable) -> bool:
    if inspect.isroutine(call):
        return inspect.iscoroutinefunction(call)
    if inspect.isclass(call):
        return False
    call_ = getattr(call, "__call__", None)  # noqa: B004
    return inspect.iscoroutinefunction(call_)


async def solve_generator_async(
    *, call: AnyCallable, stack: AsyncExitStack, sub_values: Dict[str, Any]
) -> Any:
    if is_gen_callable(call):
        cm = contextmanager_in_threadpool(contextmanager(call)(**sub_values))
    elif is_async_gen_callable(call):  # pragma: no branch
        cm = asynccontextmanager(call)(**sub_values)
    return await stack.enter_async_context(cm)


def solve_generator_sync(
    *, call: AnyCallable, stack: ExitStack, sub_values: Dict[str, Any]
) -> Any:
    cm = contextmanager(call)(**sub_values)
    return stack.enter_context(cm)


@asynccontextmanager
async def contextmanager_in_threadpool(
    cm: ContextManager[T],
) -> AsyncGenerator[T, None]:
    exit_limiter = anyio.CapacityLimiter(1)
    try:
        yield await run_in_threadpool(cm.__enter__)
    except Exception as e:
        ok = bool(
            await anyio.to_thread.run_sync(
                cm.__exit__, type(e), e, None, limiter=exit_limiter
            )
        )
        if not ok:  # pragma: no branch
            raise e
    else:
        await anyio.to_thread.run_sync(
            cm.__exit__, None, None, None, limiter=exit_limiter
        )



from typing import Any, Callable, Dict

from typing_extensions import ParamSpec, TypeAlias

P = ParamSpec("P")

AnyCallable: TypeAlias = Callable[..., Any]
DecoratedCallable: TypeAlias = AnyCallable
AnyDict: TypeAlias = Dict[str, Any]
